/*

Copyright (c) 1989, the MITRE Corporation. All rights reserved.
This software was developed by the MITRE Corporation, McLean VA.
It is distributed free of charge with the following conditions:

        1.  Redistribution and use in source and binary forms are
	    permitted provided that this copyright notice be duplicated
	    in all such forms.

        2.  MITRE will be credited should the software be used in a
            product or written about in any publications.  However,
            MITRE will not be referenced as the source in advertisements.

        3.  MITRE assumes no legal responsibility for source code and
            its subsequent use.  No warranty is expressed or implied.

        4.  If any bugs or problems are found, then they should be
            reported back to MITRE.

*/

/*

  initdev - initialize the tap device
  
 */

#include "trace.h"

#if SUN3_NIT || SUN4_NIT
	
char 	*device[5] = {"le0", "ec0", "ie0", "en0", "000"};
extern 	int quenchf;
extern u_long	snaplen;

#ifdef SUN3_NIT

/*
 * INITSOCK - socket setup for SunOS 3.x version of NIT
 */

initsock()
{
	struct	sockaddr_nit snit;
	struct	nit_ioc nioc;
	char	*strncpy();
	int	fd,i;
	
	if ((fd = socket(AF_NIT, SOCK_RAW, NITPROTO_RAW)) < 0)  {
		perror("nit socket");
		exit(-1);
	}
	
	snit.snit_family = AF_NIT;
	i=0;
	
	/* find correct ethernet device type */
	
	do {
		(void) strncpy(snit.snit_ifname, device[i], NITIFSIZ);
		i++;
	} while (bind(fd, (struct sockaddr *) &snit, sizeof(snit)) != 0 && i<5);
	
	if (i==5) {
		perror(device);
		exit (1);
	}
	
	bzero((char *) &nioc, sizeof(nioc));
	
	nioc.nioc_bufspace = NETBUFSIZ;
	nioc.nioc_chunksize = CHUNKSIZE;
	nioc.nioc_typetomatch = NT_ALLTYPES;
	
	if (snaplen) {
		nioc.nioc_snaplen = snaplen;
	}
	else 
		if (quenchf)
			nioc.nioc_snaplen = SNAPLEN + 16;
		else
			nioc.nioc_snaplen = SNAPLEN;
	
	nioc.nioc_bufalign = sizeof (int);
	nioc.nioc_bufoffset = 0;
	nioc.nioc_flags = NF_PROMISC|NF_TIMEOUT;
	nioc.nioc_timeout.tv_sec = READ_TIMEOUT;
	nioc.nioc_timeout.tv_usec = 0;
	
	if (ioctl(fd, SIOCSNIT, &nioc) != 0) {
		perror("nit ioctl");
		exit(1);
	}
	
	setuid(getuid());  /* allow process to be killed by non-super-user */
	
	return(fd);
}

#else /* SUN4_NIT */
/*
  Open a SunOS 4.0 Network Interface Tap. NIT in SunOS 4.0.0 is very
  buggy and much slower than in SunOS 3.x.  Hopefully this will be
  fixed up a little in SunOS 4.0.1.
 */
initstream()
{
	struct 	strioctl si;
	struct  timeval  timeout;
	struct 	ifreq	 ifr;
	u_long flags;
	u_int   chunksize;
	int	fd, i;
	
	/*
	  Open the NIT device, this is usually /dev/nit.
	  An exclusive open is performed as problems have been
	  encountered when more than one tcptrace is running
	  on a SUN  OS 4.0 host.
	 */
	if ((fd = open(NIT_DEV, O_RDONLY | O_EXCL)) < 0) {
		perror(NIT_DEV);
		exit(-1);
	}
	
	/*
          The documentation on this is sort of weak.  This should
	  put this stream into Message-Nondiscard mode.  I think
	  this is the proper choice?  We really don't want to
	  loose anything.
	 */
	if (ioctl(fd, I_SRDOPT, (char *) RMSGN) == -1) {
		perror("ioctl (RMSGN)");
		exit(1);
	}
	
	/*
	  Initialize the buffering module.
	 */
	if (ioctl(fd, I_PUSH, "nbuf") == -1) {
		perror("ioctl (I_PUSH nbuf)");
		exit(1);
	}
	
	si.ic_timout = 0; /* use default timeouts on all I_STR ioctl calls */
	
	/*
	  First set up timeout values
	 */
	timeout.tv_sec  = READ_TIMEOUT;
	timeout.tv_usec = 0;
	si.ic_cmd = NIOCSTIME;
	si.ic_len  = sizeof timeout;
	si.ic_dp = (char *) &timeout;
	if(ioctl(fd, I_STR, (char *)&si) == -1) {
		perror("ioctl (I_STR)");
		exit(1);
	}
	
	/*
	  Next, set up chunksize.
	 */
	si.ic_cmd = NIOCSCHUNK;
	chunksize = 0;
	si.ic_len = sizeof chunksize;
	si.ic_dp  = (char *) &chunksize;
	if(ioctl(fd, I_STR, (char *)&si) == -1) {
		perror("ioctl (I_STR)");
		exit(1);
	}
	
	/*
	  Find correct ethernet device type
	 */
	si.ic_cmd = NIOCBIND;
	si.ic_len = sizeof ifr;
	si.ic_dp  = (char *)&ifr;
	i=0;
	do {
		(void) strncpy(ifr.ifr_name, device[i], NITIFSIZ);
		ifr.ifr_name[sizeof ifr.ifr_name - 1] = ' ';
		i++;
	} while (ioctl(fd, I_STR, (char *) &si) == -1 && i<5);
	if (i==5) {
		fprintf(stderr, "No valid network interfaces found.\n");
		exit (1);
	} 
	
	/*
	  Set NIT interface flags:
	  NI_PROMISC   - put interface into promiscuous mode
	  NI_TIMESTAMP - tack a timestamp on each packet
	  NI_LEN       - indicate len of packet as received
	  NI_DROPS     - indicate the number of dropped packets
	  NI_USERBITS  - This does not seem to do anything
	 */
	flags = NI_PROMISC | NI_TIMESTAMP | NI_LEN  | NI_DROPS | NI_USERBITS;
	si.ic_cmd = NIOCSFLAGS;
	si.ic_len = sizeof flags;
	si.ic_dp = (char *)&flags;
	if(ioctl(fd, I_STR, (char *)&si) == -1) {
		perror("ioctl (I_STR)");
		exit(1);
	}
	
	/*
	  Configure snaplen
	 */
	si.ic_cmd = NIOCSSNAP;
	si.ic_len = sizeof snaplen;
	if (!snaplen) {
		if (quenchf)
			snaplen = SNAPLEN + 16;
		else
			snaplen = SNAPLEN;
	}
	si.ic_dp = (char *)&snaplen;
	if(ioctl(fd, I_STR, (char *)&si) == -1) {
		perror("ioctl (I_STR)");
		exit(1);
	}
	
	setuid(getuid());  /* allow process to be killed by non-super-user */
	return(fd);
}

#endif


#else  /* NETMON */

/*
  INITNETMON - special file setup for NETMON in BSD4.3
 */
initnetmon() 
{
	int fd;
	char *device = "/dev/netmon";
	
	if ((fd = open(device, O_RDONLY, IREAD)) < 0) {
		perror("netmon open");
		exit(-1);
	}
	return(fd);
}

#endif /* NETMON */
