/*

Copyright (c) 1989, the MITRE Corporation. All rights reserved.
This software was developed by the MITRE Corporation, McLean VA.
It is distributed free of charge with the following conditions:

        1.  Redistribution and use in source and binary forms are
	    permitted provided that this copyright notice be duplicated
	    in all such forms.

        2.  MITRE will be credited should the software be used in a
            product or written about in any publications.  However,
            MITRE will not be referenced as the source in advertisements.

        3.  MITRE assumes no legal responsibility for source code and
            its subsequent use.  No warranty is expressed or implied.

        4.  If any bugs or problems are found, then they should be
            reported back to MITRE.

*/

#include "trace.h"
#include "osi.h"

/*
 * Define the number of entries in the following structure.
 */
#define TP_TYPE_MAX TP_DT

/*
 * The tp_type structure contains 
 */ 
struct	tp_type	tp_type[TP_TYPE_MAX+1] = {
	{0, 0, "??"},
	{5, 8, "ED"},
	{5, 8, "EA"},
	{0, 0, "??"},
	{0, 0, "??"},
	{5, 8, "RJ"},		/* Not in class 4 */
	{5, 10, "AK"},
	{5, 5, "ER"},
	{7, 7, "DR"},
	{0, 0, "??"},
	{0, 0, "??"},
	{0, 0, "??"},
	{6, 6, "DC"},
	{7, 7, "CC"},
	{7, 7, "CR"},
	{5, 8, "DT"}
};

#ifdef NETMON
extern  short netmonloc;
extern 	short netmoncode;
extern  char netmonif[5];
#endif
extern 	int tot_recs;
extern 	int rawf;
extern 	int verbose;
extern 	struct conx_rec *hash_tab[TAB_SIZ];
extern  int numbers;	
extern  u_long src_host, dst_host;
extern  int host;


process_osi_packet(osi, timestamp, tot_sec, u_sec)
	u_char 	*osi;
	struct	tm *timestamp;
	long	tot_sec, u_sec;
{

	short	tp_variable_length;
	short	tp_packet_type;
	struct	clnp_fixed_header *clnp; /* pointer to start of clnp header */
	u_char	*clnp_tmp;
	u_char	*tp;		/* pointer to start of tp header */
	u_char	*addr, *osi_get_addr();
	struct	tp_fixed_header *tp_header;
	u_int	hash_src, hash_dst;
	struct	conx_rec *connect_rec, *tab_ptr, *tmp_ptr;
	struct	conx_rec *add_hash_entry();
	struct	in_addr *src_addr, *dst_addr;

	/*
	* Save pointer to CLNP header
	*/
#ifdef NETMON
	if (netmonloc == SL_IN) 
		osi = (u_char *)osi + 15;
#endif
	clnp = (struct clnp_fixed_header *) osi;

	if (verbose > 1) {
		dump_buf(clnp, osi[1]);
		dump_clnp_header(clnp);
	}

	/*
	  Make sure we've got the right protocol
	  */
	if (clnp->proto_id != 0x81) {
		if (verbose > 1) {
				printf("Unknown proto: %d\n", clnp->proto_id);
				dump_buf(clnp, 48);
			}
			return;
		}
	
#ifdef NETMON
	addr		= (u_char *)clnp+sizeof(struct clnp_fixed_header);
#else
	addr		= (u_char *)clnp+sizeof(struct clnp_fixed_header) - 1;
#endif
	dst_addr	= (struct in_addr *) osi_get_addr(addr);
	addr		= addr + addr[0] + 1;
	src_addr	= (struct in_addr *) osi_get_addr(addr);

	/* Do we want this packet. If not, throw it away */

	if (host) 
		if ((src_host == src_addr->s_addr) ||
		      (src_host == dst_addr->s_addr))
			;
		else
			return;
	else {
		if (src_host != 0) 
			if (src_host != src_addr->s_addr) 
				return;
		
		if (dst_host != 0) 
			if (dst_host != dst_addr->s_addr) 
				return;
	}

	/*
	 * Check for a segmented packet
	 */
	if (segmented(clnp)) {
		osi_log_segmented(connect_rec, *src_addr, *dst_addr,
				  tp_header, clnp, timestamp, u_sec);
		return;
	}

	/*
	  Get pointer to TP header by taking the the pointer to the
	  start of the osi header and adding the length of the clnp
	  portion of the header.
	*/
	tp = osi + osi[1];
	tp_header = (struct tp_fixed_header *) tp;
	if (verbose > 2) {
		dump_buf(tp_header, tp_header->TP_LENGTH);
		dump_tp_fixed_header(tp_header);
	}

	if ((tp_header->TP_TYPE > TP_TYPE_MAX)
	    || (tp_type[tp_header->TP_TYPE].tp_length == 0)) {
		printf("UNKNOWN TP TYPE %x\n", tp_header->TP_TYPE);
		dump_buf(clnp, 54);
		dump_buf(tp_header, tp_header->TP_LENGTH);
		dump_tp_fixed_header(tp_header);
		return;
	}
	
	/*
	 * For each packet that comes in, see if its got a hash table entry, 
	 * if not, create one, if so then point to it and the proper things
	 * will be done to it below.
	 */
	if (tp_header->TP_TYPE == TP_CR)
	   hash_dst = osi_hash(*dst_addr, *src_addr, tp_header->TP_SRC_REF, 0);
	else
	   hash_dst = osi_hash(*dst_addr, *src_addr, tp_header->TP_DST_REF, 0);

	if (hash_tab[hash_dst] == NULL) {
		/* No entry in the table for this packet, add it */
		hash_tab[hash_dst] = add_hash_entry(tp_header, *src_addr,
				      *dst_addr, tot_sec, timestamp);
		connect_rec = hash_tab[hash_dst];
	}
	else {	/* we are on the right track, scan linked list for a match */
		tab_ptr = tmp_ptr = hash_tab[hash_dst];
		while (tab_ptr != NULL) {
			if (((tp_header->TP_DST_REF == tab_ptr->th_port1) &&
			    (dst_addr->s_addr == tab_ptr->ip_addr1.s_addr)) ||
			    ((tp_header->TP_DST_REF == tab_ptr->th_port2) &&
			    (dst_addr->s_addr == tab_ptr->ip_addr2.s_addr))) {
				connect_rec = tab_ptr;
				break;
			} 
			tmp_ptr = tab_ptr;
			tab_ptr = tab_ptr->next;
		}
		if (tab_ptr == NULL) {
			tmp_ptr->next = add_hash_entry(tp_header, *src_addr,
					   *dst_addr, tot_sec, timestamp);
			connect_rec = tmp_ptr->next;
		}
	}
	
	/* at this point, connect_rec should be pointing to the a record
	 * that pertains in some way to this connection. Just update
	 * the record, print a line of output for the packet, and return */

	switch(tp_header->TP_TYPE) {

	case TP_CR:	/* Connection Request */

	/* Other than the initial negotiation stuff, there is not a whole lot
	 * of interest in this packet, for now, just print that it happened */
		
		break;

	case TP_CC:	/* Connection Confirm */

	/* This is one of the few packets that contains both the src and
	 * dst references.  Other than this, this packet dosn't contain
	 * too much info.
	 *
	 * Actions to be taken here:
	 *	1) create an entry in the hash table for this new connection
	 *	   the key here is to put two entries into the hash table,
	 *	   one for the src ref, the other for the dst ref.	 */

		/* Check the type of header format that is being used. */
		if (tp_header->TP_OPTION & 0x2) {
			connect_rec->extended = 1;
			connect_rec->flags |= EXTENDED;
		}
		else {
			connect_rec->extended = 0;
		}
		
		/* get to hash values becuase later on we will see
		   only the dst ref, not both the src and dst refs */
		
		hash_src=osi_hash(*dst_addr,*src_addr,0,tp_header->TP_SRC_REF);
		connect_rec->th_port1 = tp_header->TP_DST_REF; 
		connect_rec->th_port2 = tp_header->TP_SRC_REF; 
		if (hash_tab[hash_src] == NULL)  {
			hash_tab[hash_src] =add_hash_entry(tp_header,*src_addr,
					   *dst_addr, tot_sec, timestamp);
			connect_rec = hash_tab[hash_src];
		}
		else {
			tab_ptr = hash_tab[hash_src];
			while (tab_ptr->next != NULL) /* find end of list */
				tab_ptr = tab_ptr->next;
			tab_ptr->next =  add_hash_entry(tp_header, *src_addr,
					   *dst_addr, tot_sec, timestamp);
			connect_rec = tab_ptr->next;
		}		
		break;

	case TP_DR:	/* Disconnect Request */
		connect_rec->fin++;	/* closing */	
		break;

	case TP_DC:	/* Disconnect Confirm */
		connect_rec->closed++;	/* connection is closed */
		break;

	case TP_DT:	/* Data */
		break;

	case TP_ED:	/* Expidited Data */
		break;

	case TP_AK:	/* Acknowledgement */
		break;
	case TP_EA:	/* Expidited Ack */
		break;

	case TP_RJ:	/* Reject */
		break;
	case TP_ER:	/* Error */
		break;
	default: 
		break;
	}
	tot_recs++;
	connect_rec->packets++;
	osi_logit(connect_rec, *src_addr, *dst_addr, tp_header, clnp,
		  timestamp, u_sec);
	return;
}


/*
 * OSI hashing function, this will need some more work
 */
osi_hash(dst_addr, src_addr, dst_ref, src_ref)
	u_char dst_addr[4], src_addr[4];
	short dst_ref;
	short src_ref;
{
	int	bits;
	unsigned hash;

	hash  = (u_long) dst_addr  + (u_long) src_addr;
	hash += (1 << 31);
	bits  = dst_ref + src_ref;
	hash += (bits & 0x000007ff);
	hash  = hash % TAB_SIZ;
	return(hash);
}

/*
  This routine takes as the first argument the buffer address of the
  start of the ISO address, and as the second argument, a field to
  place the address.
*/


u_char *osi_get_addr(addr)
	struct rfc986 *addr;
{
	short idi;

	if (addr->afi != 0x47) {
		printf("AFI address format: %x not supported\n", addr->afi);
		return(0);
	}

	idi = ntohs(addr->idi);
	if (idi != 0x6) {
		printf("IDI address value: %x not supported\n", idi);
		return(0);
	}

	/*
	  Ok, we should have a DOD NSAP address at this point.  Lets
	  take a look at the version to see if it conforms to rfc 986
	  or rfc 1069.
	*/
	
	switch (addr->version) {
	case 0x1:		/* RFC 986 */
		return(&addr->ip1);
		break;

	case 0x2: 		/* RFC 1069 */
		printf("RFC 1069 addresses not yet supported\n");
		break;

	default:		/* Unknown */
		printf("UNKNOWN DOD address version\n");
		return(0);
	}
	return(0);
}


/*
  Output information about the current OSI packet
*/ 
osi_logit(connect_rec, src_addr, dst_addr, tp_header, clnp,
	  timestamp, u_sec)
	struct conx_rec *connect_rec;
        struct in_addr src_addr, dst_addr;
	struct tp_fixed_header	*tp_header;
	struct clnp_fixed_header *clnp;
	struct tm *timestamp;
	long   u_sec;
{
	int    eot;
	int    tpdu;
	u_long seq, ack;
	u_char *tp_tmp;  
	u_char *clnp_tmp;
	u_char *tp_var_header;

	printf("%02d:%02d:%02d.%02ld ", timestamp->tm_hour,
	       timestamp->tm_min, timestamp->tm_sec, u_sec/10000);

#ifdef NETMON
	switch (netmonloc) {
	case SL_IN:
		printf("SL_IN  ");
		break;
	case SL_OUT:
		printf("SL_OUT ");
		break;
	case CLNP_IN:
		printf("CL_IN  ");
		break;
	case CLNP_OUT:
		printf("CL_OUT ");
		break;
	default:
		printf("?????? ");
	}
	if (netmonif[0]) 
		printf("%s ", netmonif);
#endif
	if (src_addr.s_addr == connect_rec->ip_addr1.s_addr) 
		printf("%s", connect_rec->hostname1);
	else
		printf("%s", connect_rec->hostname2);

	switch (tp_header->TP_TYPE) { 	/* only a few have a src ref */
	case TP_CR:			/* print it when we have it  */
	case TP_CC:			/* otherwise print 0 */
	case TP_DR:
	case TP_DC:
		printf(".%d ", ntohs(tp_header->TP_SRC_REF));
		break;
	default:
		printf(".%d ", 0);
	}
	printf("-> ");
	if (src_addr.s_addr == connect_rec->ip_addr1.s_addr) 
		printf("%s", connect_rec->hostname2);
	else
		printf("%s", connect_rec->hostname1);

	printf(".%d ", ntohs(tp_header->TP_DST_REF));
	printf("len %d ", (clnp->segment_length1 << 8) + clnp->segment_length2);

	if ((tp_header->TP_TYPE == TP_DT) || (tp_header->TP_TYPE == TP_ED)) {
		seq = TP_SEQ(connect_rec->extended, tp_header);
		printf("seq# %ld ", seq);
	}

	if ((tp_header->TP_TYPE == TP_AK) || (tp_header->TP_TYPE == TP_EA)) {
		ack = TP_ACK(connect_rec->extended, tp_header);
		printf("ack# %ld ", ack);
	}

	switch (tp_header->TP_TYPE) {
	case TP_CR: 	
	case TP_CC:	
		printf("TP%d ", tp_header->TP_CLASS);
		printf("OPT=%x ", tp_header->TP_OPTION);
		/* ok to fall thru here */
	case TP_RJ:
		printf("cred: %d ", tp_header->TP_CDT);
		break;
	case TP_AK:	
		printf("cred: %d ", ntohs(TP_CREDIT(connect_rec->extended,
					      tp_header)));
		break;
	default:
		break;
	}

	printf("%s ", tp_type[tp_header->TP_TYPE].type);
	
	if (tp_header->TP_TYPE == TP_ER) {
		printf("Reject= 0x%x ", tp_header->TP_REJECT);
	}
	
	if (ce_bit_set(clnp))
		printf("CE ");
	/*
	 * Finish up here 
	 */
	switch (tp_header->TP_TYPE) {
	case TP_CR: 	
	case TP_CC:	
		break;
	case TP_DR:
		printf("REASON: 0x%x ", tp_header->TP_REASON);
	case TP_DT:
	case TP_ED:
		if (TP_IF_EOT(connect_rec->extended, tp_header) == 1)
			printf("EOT ");
		break;
	default:
		break;
	}
	tp_var_header = (u_char *)tp_header;
	tp_var_header = tp_var_header
		+ tp_type[tp_header->TP_TYPE].tp_extended_length;
	dump_tp_variable_header(tp_var_header,
				tp_header->TP_LENGTH,
				tp_type[tp_header->TP_TYPE].tp_extended_length,
				tp_header->TP_TYPE);
	printf(" OSI\n");
}

osi_log_segmented(connect_rec, src_addr, dst_addr, tp_header, clnp,
	  timestamp, u_sec)
	struct conx_rec *connect_rec;
        struct in_addr src_addr, dst_addr;
	struct tp_fixed_header	*tp_header;
	struct clnp_fixed_header *clnp;
	struct tm *timestamp;
	long   u_sec;
{
	u_char *clnp_tmp;
	
	printf("%02d:%02d:%02d.%02ld ", timestamp->tm_hour,
	       timestamp->tm_min, timestamp->tm_sec, u_sec/10000);

#ifdef NETMON
	switch (netmonloc) {
	case SL_IN:
		printf("SL_IN  ");
		break;
	case SL_OUT:
		printf("SL_OUT ");
		break;
	case CLNP_IN:
		printf("CL_IN  ");
		break;
	case CLNP_OUT:
		printf("CL_OUT ");
		break;
	default:
		printf("?????? ");
	}
#endif
	printf("%s -> ", inet_ntoa(src_addr));
	
	printf("%s SEGMENTED ", inet_ntoa(dst_addr));

	clnp_tmp = (u_char *)clnp + sizeof(struct clnp_fixed_header) - 1;
	/* move past address part and get to seg part */
	clnp_tmp = clnp_tmp + clnp_tmp[0] + 1;
	clnp_tmp = clnp_tmp + clnp_tmp[0] + 5;
	/* clnp_tmp not points to PDU length, its a short */

	printf("len %d ", (clnp_tmp[0] * 0x100 + clnp_tmp[1]));
	printf("OSI\n");
}

dump_clnp_header(clnp)
	struct clnp_fixed_header *clnp;
{
	u_char *tmp, *tmp1;
	short  i,j;

	printf("Protocol ID: %x  Header Len: %x:  Version: %x  Lifetime: %x\n",
		clnp->proto_id, clnp->length, clnp->version, clnp->lifetime);

	printf("FLAGS: Seg permitted: %x  more segs: %x  Err report: %d  TYPE: %x\n",
		clnp->segs_permitted, clnp->more_segs, clnp->err_report, clnp->type);

/*	printf("Segment length: %x%02x   Checksum: %x%02x\n",
	       clnp->segment_length1, clnp->segment_length2,
	       clnp->checksum1, clnp->checksum2);*/

	tmp = tmp1 = (u_char *)clnp + sizeof (struct clnp_fixed_header) - 1;

	printf("SRC ADDR - Len: %x AFI: %x  ", tmp[0], tmp[1]);
	for (j=2;j<=tmp[0];j++)
		printf("%02x ", tmp[j]);
	printf("\n");

	tmp = tmp + tmp[0] + 1;
	printf("DST ADDR - Len: %x AFI: %x  ", tmp[0], tmp[1]);
	for (i=2;i<=tmp[0];i++)
		printf("%02x ", tmp[i]);
	printf("\n");
	
	tmp = tmp + tmp[0] + 1;
	printf("SEG PART - DUI:  %02x%02x  Seg Offset: %02x%02x Tot Len: %02x%02x\n",
	       tmp[0], tmp[1], tmp[2], tmp[3], tmp[4], tmp[5]);

	printf("\n");
	fflush(stdout);
}



dump_tp_fixed_header(tp_header)
	struct tp_fixed_header *tp_header;
{	
	u_char *tp_var_header;

	printf("LENGTH: %x\t", tp_header->TP_LENGTH);
	printf("TYPE: %s\t", tp_type[tp_header->TP_TYPE].type);
	printf("DST: %x\t", tp_header->TP_DST_REF);
	printf("SRC: %x\n", tp_header->TP_SRC_REF);

	printf("EOT: %x\t\t", tp_header->TP_EOT);
	printf("TPDU NR: %x\t", tp_header->TP_TPDU_NR);
	printf("TU NR: %x\t", tp_header->TP_TU_NR);
	printf("REJECT: %x\n", tp_header->TP_REJECT);

	printf("CLASS: %x\t", tp_header->TP_CLASS);
	printf("OPT: %x\t\t", tp_header->TP_OPTION);
	printf("REASON: %x\n", tp_header->TP_REASON);

	tp_var_header = (u_char *)tp_header;
	tp_var_header = tp_var_header+tp_type[tp_header->TP_TYPE].tp_extended_length;
	dump_tp_variable_header(tp_var_header,
				tp_header->TP_LENGTH,
				tp_type[tp_header->TP_TYPE].tp_extended_length,
				tp_header->TP_TYPE);
	printf("\n");
}


dump_tp_variable_header(tp_var_header, max_length, fixed_length, pkt_type)
	u_char *tp_var_header;
	int max_length;
	int fixed_length;
	u_char *pkt_type;
{
	int i, length;
	u_char pcode;
	while (fixed_length < max_length) {
		pcode = *tp_var_header++;
		fixed_length++;
		length = (u_char) *tp_var_header++;
		fixed_length++;
		param_print(pcode, length, tp_var_header, pkt_type);
		fixed_length = fixed_length + length;
		tp_var_header = tp_var_header + length;
	}
}



param_print(param_code, length, value, pkt_type)
	u_char	*param_code;
	int    	length;
	u_char	*value;
	u_char	pkt_type;
{
	int j;
	
	switch(param_code) {
	case 0x0:
		break;
	case 0xc0:
		printf(" TPDU-size=");
		break;
	case 0xc1:
		printf(" TSAP-ID=");
		break;
	case 0xc2:
		printf(" TSAP-ID=");
		break;
	case 0xc3:
		return;
	case 0xc4:
		printf(" Version=");
		break;
	case 0xc5:
		printf(" Security=");
		break;
	case 0xc6:
		printf(" Addit-opt=");
		break;
	case 0xc7:
		printf(" Alt-proto-class=");
		break;
	case 0x85:
		printf(" ACK-time=");
		break;
	case 0x89:
		printf(" Throughput=");
		break;
	case 0x86:
		printf(" Error-rate=");
		break;
	case 0x87:
		printf(" Priority=");
		break;
	case 0x88:
		printf(" Transit-delay=");
		break;
	case 0x8b:
		printf(" Assign-time=");
		break;
	case 0x8c:
		printf(" Flow-ctrl-conf=");
		break;
	case 0x8a:
		printf(" Sub-seq-no=");
		break;
	default:
		printf(" UNKNOWN-PCODE(0x%x)=", param_code);
/* !!! */	length = 0;
		break;
	}
	for(j=0; j<length; j++)
		printf("%02x ", (u_char)value[j]);
	printf(" ");
}

struct conx_rec *add_hash_entry(tp_header, src_addr, dst_addr, tot_sec,
				timestamp)
	struct tp_fixed_header *tp_header;
	struct in_addr src_addr, dst_addr;
	u_long tot_sec;
	struct tm *timestamp;
{
	struct conx_rec *new_entry;
	struct hostent	*tmp_hostent;
	
	new_entry = (struct conx_rec *)malloc (sizeof (struct conx_rec));

	new_entry->fin    = new_entry->closed = new_entry->th_seq =
	new_entry->th_ack = 0;

	new_entry->extended  = 1; 		/* ASSUME Extended Mode */
	new_entry->flags |= EXTENDED;		/* ASSUME Extended Mode */

	new_entry->flags |= OSI;

	new_entry->dot_addr1 = (char *) malloc(BUFSIZ);

	strcpy(new_entry->dot_addr1, inet_ntoa(src_addr));

	new_entry->dot_addr2 = (char *) malloc(BUFSIZ);
	strcpy(new_entry->dot_addr2, inet_ntoa(dst_addr));

	new_entry->th_port1 = tp_header->TP_SRC_REF;
	new_entry->th_port2 = tp_header->TP_DST_REF;

	new_entry->hostname1 = (char *)malloc(BUFSIZ);
	new_entry->hostname2 = (char *)malloc(BUFSIZ);
	if (!numbers) {
		tmp_hostent = gethostbyaddr(&src_addr,
			      sizeof (struct in_addr), AF_INET);
		if (tmp_hostent == NULL)
			strcpy(new_entry->hostname1,new_entry->dot_addr1);
		else
			(void)strcpy(new_entry->hostname1,
				     tmp_hostent->h_name);

		tmp_hostent = gethostbyaddr(&dst_addr,
			      sizeof (struct in_addr), AF_INET);
		if (tmp_hostent == NULL)
			strcpy(new_entry->hostname2,new_entry->dot_addr2);
		else
			(void)strcpy(new_entry->hostname2,
				     tmp_hostent->h_name);
	}
	else {
		strcpy(new_entry->hostname1,new_entry->dot_addr1);
		strcpy(new_entry->hostname2,new_entry->dot_addr2);
	}
		
	new_entry->ip_addr1.s_addr = src_addr.s_addr;
	new_entry->ip_addr2.s_addr = dst_addr.s_addr;

	new_entry->timestmp 	= timestamp;
	new_entry->tot_sec  	= tot_sec;
	new_entry->packets  	= 0;
	new_entry->next     	= NULL;

	return(new_entry);
}

segmented(clnp)
	struct clnp_fixed_header *clnp;
{
	u_char *ptr;

	ptr = (u_char *)clnp + 9; /* get past the fixed part */
	ptr = ptr + ptr[0] + 1;	  /* blow past the addr part */
	ptr = ptr + ptr[0] + 1;	  /* blow past the addr part */
	if ((ptr[2] == 0) && (ptr[3] == 0))
		return(0);
	else
		return(1);		
}

ce_bit_set(clnp)
	struct clnp_fixed_header *clnp;
{
	u_char *ptr;

	ptr = (u_char *)clnp + 9; /* get past the fixed part */
	ptr = ptr + ptr[0] + 1;	  /* blow past the addr part */
	ptr = ptr + ptr[0] + 1;	  /* blow past the addr part */
	ptr = ptr + 6;		  /* move past segment part  */

	while (ptr < ((u_char*)clnp + clnp->length)) {
		if (*ptr == QOS) {
/*			printf("\nOPT = %x VALUE = %x\n", *ptr, ptr[2]);*/
			if ((ptr[2] & ~0x3f) && (ptr[2] & ~0xf7))
				return(1);
			else
				ptr = ptr + ptr[1];
		}
		else
			ptr = ptr + ptr[1];
	}
	return(0);
}




