/*

Copyright (c) 1989, the MITRE Corporation. All rights reserved.
This software was developed by the MITRE Corporation, McLean VA.
It is distributed free of charge with the following conditions:

        1.  Redistribution and use in source and binary forms are
	    permitted provided that this copyright notice be duplicated
	    in all such forms.

        2.  MITRE will be credited should the software be used in a
            product or written about in any publications.  However,
            MITRE will not be referenced as the source in advertisements.

        3.  MITRE assumes no legal responsibility for source code and
            its subsequent use.  No warranty is expressed or implied.

        4.  If any bugs or problems are found, then they should be
            reported back to MITRE.

*/

/*
  Define the structure for a CLNP fixed header, variable length
  headers follow the fixed part.
*/

struct clnp_fixed_header {
	u_char	proto_id;
	u_char	length;
	u_char	version;
	u_char	lifetime;
	u_char	segs_permitted:1,
		more_segs:1,
		err_report:1,
 		type:5;
	u_char 	segment_length1;  
	u_char 	segment_length2;  
	u_char	checksum1;
	u_char	checksum2;
};

/*
  Define the structures for the different address formats.
*/


struct rfc986 {		/* Yes this is obsolete */
	u_char	length;
	u_char	afi;		/* Athority and Format Identifier */
	u_short	idi;	
	u_char	version;	/* Version number */
	u_char	ip1;
	u_char	ip2;
	u_char	ip3;
	u_char	ip4;
	u_char	ipupid;		/* IP user protocol ID */
};

/*
 * OSI TP4 packet types
 */
#define TP_ED	1
#define TP_EA	2
#define TP_RJ	5
#define TP_AK	6
#define TP_ER	7
#define TP_DR	8
#define TP_DC	12
#define TP_CC	13
#define TP_CR	14
#define TP_DT	15

struct tp_fixed_header {
	/* Byte 1, header length. Same for all packet types */
	u_char	tp_length;

	/* Byte 2, Code and CDT signal */
#ifdef NETMON
	u_char	tp_cdt:4,
		tp_type:4;
#else	
	u_char	tp_type:4,
		tp_cdt:4;
#endif
	/* Bytes 3 and 4.  The transport connection identity at the
	 * remote connection.  Same for all packet types.	 */
	u_short	tp_dst_ref;

	/* bytes 5 - 8 */

	union {
		struct tp_src_struct {
			/* 5 and 6 */
			union {
				u_short tp_src_ref;
				struct {
#ifdef NETMON
					u_char	tp_tpdu_nr:7;
					u_char	tp_eot:1;
#else
					u_char	tp_eot:1;
					u_char	tp_tpdu_nr:7;
#endif					
					u_char	filler;
				} tp_dt_struct;
				struct {
					u_char	tp_tu_nr;
					u_char	filler;
				} tp_ak_struct;
				struct {
					u_char	tp_reject;
					u_char	filler;
				} tp_er_struct;
			} tp_src;

			/* 7 and 8 */
			union {
				struct {
#ifdef NETMON
					u_char	tp_option:4;
					u_char 	tp_class:4;
#else
					u_char 	tp_class:4;
					u_char	tp_option:4;
#endif
					u_char 	filler;
				} tp_con_struct;
				u_char	tp_reason;
			} tp_cl;
		} tp_normal;

		/* 5,6,7 and 8 */
		union {
			struct {
#ifdef NETMON
				u_int tp_tpdu_nr:31;
				u_char tp_eot:1;
#else
				u_char tp_eot:1;
				u_int tp_tpdu_nr:31;
#endif
			} tp_extended;
			struct {
				u_int tp_ack:32;
			} tp_ack_extended;
		} end_union;
	} tp_mode;
	u_short	tp_x_cdt;
};

#define TP_LENGTH	tp_length
#define TP_TYPE		tp_type
#define TP_CDT		tp_cdt
#define TP_DST_REF	tp_dst_ref
#define TP_SRC_REF	tp_mode.tp_normal.tp_src.tp_src_ref
#define TP_EOT		tp_mode.tp_normal.tp_src.tp_dt_struct.tp_eot
#define TP_TPDU_NR 	tp_mode.tp_normal.tp_src.tp_dt_struct.tp_tpdu_nr
#define TP_TU_NR	tp_mode.tp_normal.tp_src.tp_ak_struct.tp_tu_nr
#define TP_REJECT	tp_mode.tp_normal.tp_src.tp_er_struct.tp_reject
#define TP_CLASS	tp_mode.tp_normal.tp_cl.tp_con_struct.tp_class
#define TP_OPTION	tp_mode.tp_normal.tp_cl.tp_con_struct.tp_option
#define TP_REASON	tp_mode.tp_normal.tp_cl.tp_reason

#define TP_X_CDT	tp_x_cdt
#define TP_X_EOT	tp_mode.end_union.tp_extended.tp_eot
/*#define TP_X_TPDU_NR	tp_mode.end_union.tp_extended.tp_tpdu_nr*/
#define TP_X_TPDU_NR	tp_mode.end_union.tp_ack_extended.tp_ack  /* need all 32 bits, yes this looks weird */
#define TP_X_TU_NR	tp_mode.end_union.tp_ack_extended.tp_ack

#define TP_ACK(format, ptr) ((format) ? \
			     ntohl((ptr)->TP_X_TU_NR) : \
			     (ptr)->TP_TU_NR)

#define TP_CREDIT(format, ptr) ((format) ? \
				(ptr)->TP_X_CDT : \
				(ptr)->TP_CDT)

#define TP_SEQ(format, ptr) ((format) ? \
			     ntohl((ptr)->TP_X_TPDU_NR) : \
			     (ptr)->TP_TPDU_NR)

#define TP_IF_EOT(format, ptr) ((format) ? \
				(ptr)->TP_X_EOT : \
				(ptr)->TP_EOT)

struct tp_type {
	/*
	 * header length in bytes of a normal TP4 header.
	 */
	short	tp_length;
	/*
	 * Header length in bytes of an extended TP4 header.
	 */
	short	tp_extended_length;
	/*
	 * An ASCII (printable) TP4 packet type.
	 */
	char	*type;
};

/*
 * connection record for use in the OSI hash table.
 */
struct	osi_conx_rec {
	u_long	seq;
        u_long  ack;
	char	*hostname1;
	char	*hostname2;
	char	*dot_addr1;
	char	*dot_addr2;
	struct 	in_addr ip_addr1;
	struct 	in_addr ip_addr2;
	u_long	fin;
	struct	tm *timestmp;
	long	tot_sec;
	u_short	src_ref;
	u_short	dst_ref;
	short	closed;
	int	packets;
	struct	osi_conx_rec *next;
};

#define QOS	0xC3
