/*

Copyright (c) 1989, the MITRE Corporation. All rights reserved.
This software was developed by the MITRE Corporation, McLean VA.
It is distributed free of charge with the following conditions:

        1.  Redistribution and use in source and binary forms are
	    permitted provided that this copyright notice be duplicated
	    in all such forms.

        2.  MITRE will be credited should the software be used in a
            product or written about in any publications.  However,
            MITRE will not be referenced as the source in advertisements.

        3.  MITRE assumes no legal responsibility for source code and
            its subsequent use.  No warranty is expressed or implied.

        4.  If any bugs or problems are found, then they should be
            reported back to MITRE.

*/

/*

  tcpip.c contains a set of TCP/IP specific tracing routines.

 */  

#include "trace.h"

/*
  Externally defined variables
*/

extern short numbers, retain;

extern struct conx_rec *list_ptr, *hash_tab[TAB_SIZ];

extern int numdropped, numseen, quenchf;

extern int rawf;
extern u_long me;
extern struct conx_rec *connect_rec;

extern short tot_recs;	 /* total number of records in use */
char *malloc();
/*
 * LOGIT - write header info
 */
void logit(con_rec, seq_num, time, u_sec, len, close_flag, tackf, ack_num,
      ack2, sw,window, mustswap, flags, optlen, optstr)

	/*
	  NETMON: TCP option logging, and swaplen parameter for ip length.
	  */

	struct	conx_rec *con_rec;
	u_long	seq_num, ack_num, ack2;
	long	u_sec;
	struct	tm *time;
	short	len, close_flag,window;
	u_char	flags;
	u_char	*optstr;
	int	optlen;
	int	tackf,sw; 
	short	mustswap;
{
	char	*finn	= "F";
	char	*syn	= "S";
	char	*rst	= "R";
	char	*push	= "P";
	char	*ack	= "A";
	char	*urg	= "U";
	char	*blank	= " ";
	
	int i;
	
	if (!sw) {
		(void) printf("%02d:%02d:%02d.%02ld  %s.%u > %s.%u seq# %lu len %d ack# %lu win: %u %s %s %s %s %s %s  %d ",
			      time->tm_hour, time->tm_min, time->tm_sec, u_sec/10000,
			      numbers ? con_rec->dot_addr1 : con_rec->hostname1,
			      con_rec->th_port1,
			      numbers ? con_rec->dot_addr2 : con_rec->hostname2,
			      con_rec->th_port2, 
			      seq_num - con_rec->th_seq,
			      (mustswap > 0) ? ntohs(len) : len,
			      tackf ? ack_num - con_rec->th_ack : ack_num-ack_num,
			      (u_short)(window),
			      (flags &	TH_ACK) ? ack : blank,
			      (flags &	TH_FIN) ? finn : blank,
			      (flags &	TH_SYN) ? syn : blank,
			      (flags &	TH_RST) ? rst : blank,
			      (flags &	TH_PUSH) ? push : blank,
			      (flags &	TH_URG) ? urg : blank, optlen);
	}
	else {
		(void) printf("%02d:%02d:%02d.%02ld %s.%u > %s.%u seq# %lu len %d ack# %lu win: %u %s %s %s %s %s %s  %d ",
			      time->tm_hour,time->tm_min,time->tm_sec,u_sec/10000,
			      numbers ? con_rec->dot_addr2: con_rec->hostname2,
			      con_rec->th_port2,
			      numbers ? con_rec->dot_addr1 : con_rec->hostname1,
			      con_rec->th_port1,
			      seq_num - con_rec->th_ack,
			      /*seq_num-con_rec->th_seq - 1,*/
			      (mustswap > 0) ? ntohs(len) : len,
			      tackf ? ack_num - con_rec->th_seq : ack2-ack2,
			      (u_short)(window),
			      (flags &	TH_ACK) ? ack : blank,
			      (flags &	TH_FIN) ? finn : blank,
			      (flags &	TH_SYN) ? syn : blank,
			      (flags &	TH_RST) ? rst : blank,
			      (flags &	TH_PUSH) ? push : blank,
			      (flags &	TH_URG) ? urg : blank, optlen);
		
		
	}
	
	if(optlen) for(i =0 ; i < optlen; i++)
		printf("%02x ", optstr[i]);
	
	if (close_flag == 1) {
		(void) printf(" CLOSE\n");
		fflush(stdout);
	}
	else if (close_flag == 2)
		(void) printf(" NEW\n");
	else (void) printf("\n");
	
	return;
}


/*
 * TCP_HASH - hash function
 */

tcp_hash(ipheader, tcpheader)
	struct	ip *ipheader;
	struct	tcphdr *tcpheader;
{
	int	bits;
	unsigned hash;
	
	hash = ipheader->ip_src.s_addr + ipheader->ip_dst.s_addr;
	hash += (1 << 31);
	bits = tcpheader->th_sport + tcpheader->th_dport;
	hash += (bits & 0x000007ff);
	hash %= TAB_SIZ;
	return(hash);
	
}


/*
 * SET_TCP_IP - set tcp/ip header pointers (contains sun specific code)
*/

set_tcp_ip(tcpheader, ipheader)
	struct	tcphdr **tcpheader;
	struct	ip **ipheader;
{
	struct	icmp *icmpheader;
	
	struct	conx_rec *qtab_ptr;
	static	unsigned hash;
	static	char *inet_ntoa(), *qsource, *qdestin;
	
#ifndef NIT_CATCH /* NETMON */
	int mustswap = 0;
#endif
	
/*	printf(">set_tcp_ip\n");*/
	/* check for a source quench */
	if ((quenchf) && ((*ipheader)->ip_p == IPPROTO_ICMP)) {
		
		if ((*ipheader)->ip_len < ((2 * sizeof(struct ip *)) +
					   sizeof(struct icmp *) + 8)) {
			printf("Bad ICMP message\n");
			return(0);
		}
		icmpheader = (struct icmp *) (((u_int) (*ipheader) +
					       (u_int)(((*ipheader)->ip_hl) * 4)));
		if (((icmpheader)->icmp_type) == ICMP_SOURCEQUENCH) {
			
			printf("Source Quench by %s for -> ",
			       inet_ntoa((*ipheader)->ip_src));
			*ipheader = (struct ip *) (((u_int)(&(icmpheader->icmp_ip))
						    + sizeof(int) - 1) & ~(sizeof(int) - 1));
			*tcpheader = (struct tcphdr *) ((u_int) *ipheader +
							(u_int) (*ipheader)->ip_hl * 4);
			/*
			 * NETMON: port to VAX, must convert to DEC byte order.
			 * The network to host order macros are noops in SUN,
			 * so need not ifdef this bit.
			 */
			(*tcpheader)->th_sport = ntohs((*tcpheader)->th_sport);
			(*tcpheader)->th_dport = ntohs((*tcpheader)->th_dport);
			(*tcpheader)->th_seq = ntohl((*tcpheader)->th_seq);
			
			hash = tcp_hash(*ipheader, *tcpheader);
			if ((hash_tab[hash]) != NULL) {
				qtab_ptr = hash_tab[hash];
				while ((qtab_ptr != NULL) &&
				       !(((*ipheader)->ip_src.s_addr == qtab_ptr->ip_addr1.s_addr) &&
					 ((*tcpheader)->th_sport == qtab_ptr->th_port1) &&
					 ((*ipheader)->ip_dst.s_addr == qtab_ptr->ip_addr2.s_addr) &&
					 ((*tcpheader)->th_dport == qtab_ptr->th_port2)) &&
				       !(((*ipheader)->ip_src.s_addr == qtab_ptr->ip_addr2.s_addr) &&
					 ((*tcpheader)->th_sport == qtab_ptr->th_port2) &&
					 ((*ipheader)->ip_dst.s_addr == qtab_ptr->ip_addr1.s_addr) &&
					 ((*tcpheader)->th_dport == qtab_ptr->th_port1))) {
					
					qtab_ptr = qtab_ptr->next;
				}
			}
			qsource = malloc(15);
			qdestin = malloc(15);
			if ((strlen(strcpy(qsource,inet_ntoa((*ipheader)->ip_src))))<2) {
				printf("unknown source\n");
				return(0);
			}
			if ((strlen(strcpy(qdestin,inet_ntoa((*ipheader)->ip_dst))))<2) {
				printf("unknown destination\n");
				return(0);
			}
			printf("%s.%u > %s.%u seq# %lu \n", 
			       qsource,
			       (*tcpheader)->th_sport,
			       qdestin,
			       (*tcpheader)->th_dport,
			       qtab_ptr==NULL ? (*tcpheader)->th_seq :
			       (*tcpheader)->th_seq - qtab_ptr->th_seq);
		}
		
		return(0);
	}
	
	/* otherwise log only TCP packets */
	
/*	printf("---in set_tcp_ip\n");
	dump_ip_header(*ipheader);
	dump_buf((*ipheader), 32); */
	if ((*ipheader)->ip_p != IPPROTO_TCP)  {
		
#ifdef NETMON_DEBUG
		printf("ip_p == %d\n",(*ipheader)->ip_p);
#endif
/*		printf("UNKNOWN IP HEADER TYPE, %d\n", (*ipheader)->ip_p);*/
		return(0);
	}
	
	/*
	 * the u_int casts fail on brain-dead machines.
	 * multiply by 4 as ip_hl is in 32-bit words.
	 */
	
	*tcpheader = (struct tcphdr *) ((u_int) *ipheader +
					(u_int) (*ipheader)->ip_hl * 4);
	
	/*
	 * NETMON:  switch from network to host order for code on VAX.
	 * The hton, ntoh routines are noops in SUN.
	 */
	(*ipheader)->ip_len = ntohs((*ipheader)->ip_len);
	(*tcpheader)->th_sport = ntohs((*tcpheader)->th_sport);
	(*tcpheader)->th_dport = ntohs((*tcpheader)->th_dport);
	(*tcpheader)->th_seq = ntohl((*tcpheader)->th_seq);
	(*tcpheader)->th_ack = ntohl((*tcpheader)->th_ack);
	(*tcpheader)->th_win = ntohs((*tcpheader)->th_win);
	
/*	printf("<set_tcp_ip\n");*/
#ifdef NIT_CATCH
	return(1);
#else /* NETMON */
	return(mustswap);
#endif
	
}


/*
 * GETVAL - main processing for tracing
 */

void process_ip_packet(src_host, dst_host, host, src_port, dst_port,
		       port, timestamp, tot_sec, u_sec, ipheader)

	u_long	src_host, dst_host;
	u_short	src_port, dst_port;
	int	host, port;
	struct	tm *timestamp;
	long	tot_sec, u_sec;
	struct	ip *ipheader;
/*	unsigned char ip_header;*/	/* start of this ip packet */
{
	struct	tcphdr *tcpheader;
	
	int	fin;
	int	ackf;
	int	switchack = 0;
	unsigned hash;
	short	found_src, found_port;
	struct	hostent *tmp_hostent;
	struct	conx_rec *tab_ptr, *tmp_ptr;
	long	ack_num, ack_num2;
	
	struct	 tm *localtime();
	struct	 hostent *gethostbyaddr();
	char	 *strcpy(), *inet_ntoa();
	void	 get_buf();
	
	/*
	 * NETMON:  additional variables for logging TCP options.
	 */
	char *tp, optstr[60];
	int optlen;
	
/*
  The following commented out by gregh.
  Several parts of the code below are dependent on
  the netmonswap variable being defined.
  */

/* #ifndef NIT_CATCH */
	
	int netmonswap = 0;
	
/* #endif */
	
/*	printf(">process_ip_packet\n");*/
	found_src = found_port = 0;
			
	ackf = 0;
			
	/*
	 * get the tcpheader, ipheader and timestamp set, with
	 * bp in the appropriate place (get a packet).
	 */
	
/*		dump_buf(ipheader, 32);*/
#ifdef NIT_CATCH
	if (!set_tcp_ip(&tcpheader, &ipheader)) {
		return;
	}
#else /* NETMON */
	    if ((netmonswap = set_tcp_ip(&tcpheader, &ipheader)) == 0) {
		return;
	}
#endif
	    /* is this an ack ???? */
	    
	    if (tcpheader->th_flags & TH_ACK) {
		    ackf = 1;
		    ack_num = tcpheader->th_ack;
	    }
	    
	    /* do we want this packet? */
			
	    if (src_host != 0) {
		    if ((src_host != ipheader->ip_src.s_addr) && 
			!(ackf && src_host == ipheader->ip_dst.s_addr)) {
			    if (!host) {
				    return;
			    }
		    } else
			    found_src = 1;
	    }
	    if (dst_host != 0) {
		    if ((dst_host != ipheader->ip_dst.s_addr) &&
			((!host && !ackf)||(host && !found_src) || 
			 (ackf && !found_src && dst_host !=
			  ipheader->ip_src.s_addr))) {
			    return;
		    }
	    }
	    if (src_port != 0) {
		    if ((src_port != tcpheader->th_sport) && 
			!(ackf && src_port==tcpheader->th_dport)) {
			    if (!port)  {
				    return;
			    }
		    } else {
			    found_port= 1;
		    }
	    }
	    if (dst_port != NULL) {
		    if ((dst_port != tcpheader->th_dport)
			&& ((!port && !ackf)
			    || (port && !found_port)
			    || (ackf && !found_port && dst_port !=
				tcpheader->th_sport))) {
			    return;
		    }
	    }
	    
	    /*
	     * NETMON:  does it have TCP options to be logged?
	     */
	    tcpheader->th_off -= 5;
	    optlen = (int)(tcpheader->th_off * 4);
	    if (optlen) {
		    int i;
		    
		    tp = (char *)tcpheader + sizeof(struct tcphdr);
		    bcopy(tp, optstr, optlen);
#ifdef NETMON_DEBUG
		    printf("nxt seg's options: ");
		    for(i =0 ; i < optlen; i++)
			    printf("%02x ", optstr[i]);
		    printf("\n");
#endif
	    }
	    
	    /* we want it.	do we already have it? */
	    
	    switchack = 0;
	    
	    hash = tcp_hash(ipheader, tcpheader);
	    if ((hash_tab[hash]) != NULL) {
		    tab_ptr = tmp_ptr = hash_tab[hash];
		    while ((tab_ptr != NULL) &&
			   !((ipheader->ip_src.s_addr == tab_ptr->ip_addr1.s_addr) &&
			     (tcpheader->th_sport == tab_ptr->th_port1) &&
			     (ipheader->ip_dst.s_addr == tab_ptr->ip_addr2.s_addr) &&
			     (tcpheader->th_dport == tab_ptr->th_port2)) &&
			   !((ipheader->ip_src.s_addr == tab_ptr->ip_addr2.s_addr) &&
			     (tcpheader->th_sport == tab_ptr->th_port2) &&
			     (ipheader->ip_dst.s_addr == tab_ptr->ip_addr1.s_addr) &&
			     (tcpheader->th_dport == tab_ptr->th_port1))) {
			    tmp_ptr = tab_ptr;
			    tab_ptr = tab_ptr->next;
		    }
				
		    if (tab_ptr != NULL) {
			    /*
			     * found the connection.  if it is
			     * closing check to see if both sides
			     * have sent fins, and close if so.
			     */
			    
			    if ( ((ipheader->ip_src.s_addr==tab_ptr->ip_addr2.s_addr) &&
				  (tcpheader->th_sport==tab_ptr->th_port2)) ||
				((ipheader->ip_dst.s_addr==tab_ptr->ip_addr1.s_addr) &&
				 (tcpheader->th_dport==tab_ptr->th_port1)) ) {
				    switchack = 1;
				    if ((tcpheader->th_flags & TH_SYN)) {
					    if (!rawf)
						    tab_ptr->th_ack = tcpheader->th_seq;
					    logit(tab_ptr,tcpheader->th_seq,timestamp,
						  u_sec,ipheader->ip_len, 0, ackf,
						  ack_num, ack_num2,
						  switchack, tcpheader->th_win,
						  netmonswap, 
						  tcpheader->th_flags,
						  optlen, optstr); /* NETMON */
					    tab_ptr->th_ack++;
					    switchack = 0;
					    tab_ptr->packets++;
					    return;
				    }
			    }
			    
			    if (tcpheader->th_flags & TH_FIN) {
				    /*printf("connection closing --\n");*/
				    fin = ipheader->ip_src.s_addr - ipheader->ip_dst.s_addr;
				    if (fin == 0) {
					    /*
					     * there's a loop, so
					     * assign the fin the
					     * difference in ports
					     */
					    if (tcpheader->th_sport == tab_ptr->th_port1)
						    tab_ptr->fin = tab_ptr->th_port1 - tab_ptr->th_port2;
					    else
						    tab_ptr->fin = tab_ptr->th_port2 - tab_ptr->th_port1;
				    }
				    
#ifdef NIT_CATCH /* SUN (vs. NETMON) won't see fin on own outbound */
						
				    if (ipheader->ip_dst.s_addr == me) {
					    /* since I'm a sun, I
					     * can't see my own FINs
					     * going out,
					     */
					    logit(tab_ptr, tcpheader->th_seq, timestamp,
						  u_sec, ipheader->ip_len, 1,ackf,ack_num,
						  ack_num2,switchack,
						  tcpheader->th_win, 
						  netmonswap, 
						  tcpheader->th_flags,
						  optlen, optstr);
					    tab_ptr->closed++;
					    tab_ptr->packets++;
					    return;
				    }
				    
#endif
				    if (tab_ptr->fin == 0) {
					    tab_ptr->fin = fin;
					    logit(tab_ptr, tcpheader->th_seq, timestamp,
						  u_sec, ipheader->ip_len, 0,ackf,ack_num,
						  ack_num2,switchack,
						  tcpheader->th_win,
						  netmonswap, 
						  tcpheader->th_flags,optlen, optstr);
					    return;
				    } else if (fin != tab_ptr->fin) {
					    logit(tab_ptr, tcpheader->th_seq, timestamp,
						  u_sec, ipheader->ip_len, 1,ackf,ack_num,
						  ack_num2,switchack,
						  tcpheader->th_win,
						  netmonswap, 
						  tcpheader->th_flags,
						  optlen, optstr);
					    tab_ptr->closed++;
					    tab_ptr->packets++;
					    return;
				    }
			    } else if (tcpheader->th_flags & TH_RST) {
				    logit(tab_ptr,tcpheader->th_seq,
					  timestamp,
					  u_sec, ipheader->ip_len, 1,ackf,
					  ack_num,ack_num2,switchack,
					  tcpheader->th_win,
					  netmonswap, 
					  tcpheader->th_flags,
					  optlen, optstr);
				    tab_ptr->closed++;
				    tab_ptr->packets++;
				    return;
			    } else {
				    
				    /*
				     * Not closing -- update the
				     * time and logit.
				     */
				    logit(tab_ptr, tcpheader->th_seq, timestamp,
					  u_sec, ipheader->ip_len, 0,ackf,
					  ack_num,ack_num2,switchack,
					  tcpheader->th_win,
					  netmonswap, 
					  tcpheader->th_flags,
					  optlen, optstr);
				    if (switchack) switchack = 0;
				    tab_ptr->packets++;
				    return;
			    }
		    }
	    }
	    
	    /*
	     * We don't have a record for this connection.	So we
	     * get all the appropriate information for the record.
	     * fancy dancing with strcpy done because
	     * gethostbyaddr and inet_ntoa use a static area to do
	     * work -- we cant point there.
	     */
	    connect_rec = (struct conx_rec *) malloc (sizeof(struct conx_rec));
	    connect_rec->next = NULL;
	    connect_rec->fin = connect_rec->closed = 0;
	    connect_rec->dot_addr1 = (char *) malloc (BUFSIZ);
	    (void) strcpy(connect_rec->dot_addr1,inet_ntoa(ipheader->ip_src));
	    connect_rec->dot_addr2 = (char *) malloc (BUFSIZ);
	    (void) strcpy(connect_rec->dot_addr2,
			  inet_ntoa(ipheader->ip_dst));
	    if (!numbers) {
		    connect_rec->hostname1 = malloc(BUFSIZ);
		    tmp_hostent = gethostbyaddr((char *)&ipheader->ip_src,
						sizeof(struct in_addr), AF_INET);
		    if (tmp_hostent == NULL)
			    strcpy(connect_rec->hostname1,connect_rec->dot_addr1);
		    else
			    (void)strcpy(connect_rec->hostname1,
					 tmp_hostent->h_name);
	    }
	    if (!numbers) {
		    connect_rec->hostname2 =
			    (char *) malloc(BUFSIZ);
		    tmp_hostent =
			    gethostbyaddr((char *)&ipheader->ip_dst,
					  sizeof(struct in_addr), AF_INET);
		    if (tmp_hostent == NULL)
			    strcpy(connect_rec->hostname2,
				   connect_rec->dot_addr2);
		    else 
			    (void) strcpy(connect_rec->hostname2,
					  tmp_hostent->h_name);
	    }
	    connect_rec->th_seq = ((rawf || (!(tcpheader->th_flags & 
					    TH_SYN))) ? 0 : tcpheader->th_seq+1);
	    connect_rec->th_ack = ((rawf || (!(tcpheader->th_flags &
					    TH_SYN))) ? 0 : tcpheader->th_ack+1);
	    connect_rec->th_port1 = tcpheader->th_sport;
	    connect_rec->th_port2 = tcpheader->th_dport;
	    connect_rec->ip_addr1.s_addr = ipheader->ip_src.s_addr;
	    connect_rec->ip_addr2.s_addr = ipheader->ip_dst.s_addr;
	    connect_rec->timestmp = timestamp;
	    connect_rec->tot_sec = tot_sec;
	    
	    connect_rec->packets = 1;
	    if ((hash_tab[hash]) == NULL)
			hash_tab[hash] = connect_rec;
	    else
			tmp_ptr->next = connect_rec;
	    tot_recs++;
	    switchack = 0;
	    if (!rawf) {
		    tcpheader->th_ack++;
		    tcpheader->th_seq++;
	    }
	    logit(connect_rec, tcpheader->th_seq, timestamp,
		  u_sec, ipheader->ip_len, 2, ackf,
		  tcpheader->th_ack,
		  ack_num2,switchack,tcpheader->th_win,
		  netmonswap,
		  tcpheader->th_flags,
		  optlen, optstr);
}

dump_ip_header(ipheader)
	struct ip *ipheader;
{
	printf("Header len: %d\t\tVersion: %d\n", ipheader->ip_hl, ipheader->ip_v);
	printf("Type Of Ser: %d\t\t IP len: %d\n", ipheader->ip_tos, ipheader->ip_hl); 
	printf("IP id: %d\t\tIP offset: %d\n", ipheader->ip_id, ipheader->ip_off);
	printf("TTL: %d\t\tProtocol: %d\n", ipheader->ip_ttl, ipheader->ip_p);

	printf("Ip Src: %s\t\tIP Dst: %s\n", inet_ntoa(ipheader->ip_src),
	        inet_ntoa(ipheader->ip_dst));
}


