static char rcsid[] = "$Header: /data/src/local/pdutrace/RCS/trace.c,v 1.1 89/11/04 20:53:53 root Exp $";
/*

Copyright (c) 1989, the MITRE Corporation. All rights reserved.
This software was developed by the MITRE Corporation, McLean VA.
It is distributed free of charge with the following conditions:

        1.  Redistribution and use in source and binary forms are
	    permitted provided that this copyright notice be duplicated
	    in all such forms.

        2.  MITRE will be credited should the software be used in a
            product or written about in any publications.  However,
            MITRE will not be referenced as the source in advertisements.

        3.  MITRE assumes no legal responsibility for source code and
            its subsequent use.  No warranty is expressed or implied.

        4.  If any bugs or problems are found, then they should be
            reported back to MITRE.

*/

/*
 * tcptrace.c:  TCP connection monitoring.
 * By: K. Thompson
 *
 * Port to NETMON: A. Mankin
 * Port to SunOS 4.0: G. Hollingsworth
 * Restructuring, and addition of OSI:  G. Hollingsworth
 *
 */

#include "trace.h"

/*
 * NETMON:  NIT_CATCH is undefined for port to BSD4.3
 */

/*
  Definitions for runtime options
 */
short	numbers	= 0;	/* print formatting (dot address vs hostname) */
short	retain	= 0;	/* keep track of closed connections */
int	count	= 0;
int	host	= 0;
int	port	= 0;
int	rawf 	= 0;
int	quenchf	= 0;
int	tracetype	= 0;  
u_long	snaplen		= 0;
short 	tot_recs;	 /* total number of records in use */

short	sleepf	= 0, milhrs, milmins, milsecs, sleepsecs;

u_long 	src_host, dst_host;
u_short	src_port, dst_port;
struct 	hostent	  *tmp_hostent;

#ifdef SUN4_NIT
int	readsize;
struct 	strioctl si;
int	first_read = 0;
#endif

#ifdef PERFDEBUG
int	old_sec_dropped 	= 0;
long	last_second		= 0;
int	old_dropped 		= 0;
int	packets_this_second 	= 0;
int	pkt_offset 		= 0;
struct	timeval	testtime, oldtime;
struct	timezone testzone;
#endif

char	*malloc(), *realloc(), *wbuf;
void	cleanup1();
u_long	inet_addr();

#ifdef NIT_CATCH

unsigned char buf[CHUNKSIZE+2048];

#else /* NETMON:  slop is not useful. */

short	netmoncode 	= 0;
short	netmonloc 	= 0;
char 	netmonif[5];
unsigned char buf[CHUNKSIZE];

#endif

int	fd;
int     numdropped = 0;
int     numseen = 0;
int	verbose = 0;

#define TCPTRACE 1
#define OSITRACE 2	

u_long	me;		/* this machine's address. */
	
struct	conx_rec *hash_tab[TAB_SIZ], *connect_rec, *list_ptr = NULL;
	
#ifdef NETMON	
	struct netmonstr *nt; /* NETMON */
#endif
	

/*
 * GET_BUF - get a buffer of headers (contains sun specific code)
 */

void get_buf(fd, bp, endbuf)
	int	 fd;
	unsigned char **bp, **endbuf;
{

#ifdef SUN4_NIT
	static int old_dropped = 0;
#else	
	static int old_seq = -1;
#endif
	
int	  diff, n, tryagain;
	
#ifdef NIT_CATCH
	
#ifdef SUN4_NIT
	int  chunksize;
	struct nit4_hdr *nheader;
#else
	struct nit_hdr *nheader;
#endif
#endif
	void	  cleanup();
	*bp 	= NULL;
	*endbuf = NULL;
	
	do {
#ifdef PERFDEBUG
		printf("Packets in this buffer = %d\n\n", pkt_offset);
#endif
#ifdef SUN4_NIT
		readsize = READSIZE;
		if ((n = read(fd, buf, readsize)) < 0) {
#else
		if ((n = read(fd, buf, sizeof (buf))) < 0) {
#endif
			(void) cleanup();
			perror("read buffer");
			exit(-1);
		}
#ifdef SUN4_NIT
		/*
		  Next, set up chunksize.
		  */
		if (first_read == 0) {
			first_read++;
			si.ic_cmd = NIOCSCHUNK;
			chunksize = CHUNKSIZE;
			si.ic_len = sizeof chunksize;
			si.ic_dp  = (char *) &chunksize;
			if(ioctl(fd, I_STR, (char *)&si) == -1) {
				perror("ioctl (I_STR)");
				exit(1);
			}
		}
#endif
#ifdef PERFDEBUG

		gettimeofday(&testtime, &testzone); 
		pkt_offset=0;
		printf("Read %d bytes time: %ld.%ld elapse: %ld.%ld\n",
		       n, testtime.tv_sec,
		       testtime.tv_usec,
		       testtime.tv_sec - oldtime.tv_sec,
		       testtime.tv_usec - oldtime.tv_usec);
		oldtime.tv_sec  = testtime.tv_sec;
		oldtime.tv_usec = testtime.tv_usec;
#endif
#ifdef NIT4_DEBUG
		printf("got %d bytes\n", n);
#endif
#ifdef NETMON_DEBUG
		printf(" netmon read %d\n", n); fflush(stdout);
#endif
			
#ifdef NIT_CATCH
#ifdef SUN4_NIT
		nheader = (struct nit4_hdr *) buf; 
#ifndef PERFDEBUG
		if ((diff = nheader->DROPPED - old_dropped) > 0)
			(void)printf("dropped %d nit packets\n", diff);
			
		old_dropped = nheader->DROPPED;
#endif
			
	/*
	  I'm not sure wether the following is correct.  The line below
	  that is commented out does a similar things at the SUN3_NIT
	  version does, however, I think that the SUN3_NIT form is incorrect.
	  See set_tcp_ip() for details.
         */
		*bp = buf;
		/*    *bp = buf + sizeof(struct nit4_hdr); */
		*endbuf = buf + n;
		tryagain = 0;
#else /* SUN3_NIT */
		nheader = (struct nit_hdr *) buf; 
		
		switch(nheader->nh_state)  {
				
		case NIT_CATCH:
		case NIT_SEQNO:
			/* check for dropped nit packets */
			
			if ((diff = nheader->nh_seqno - old_seq) > 1)
				(void)printf("dropped %d nit packets\n", diff);
			
			old_seq += diff;
			
			
			/* set bp and endbuf for packet retrieval */
			
			*bp = buf + sizeof(struct nit_hdr);
			*endbuf = buf + n;
			tryagain = 0;
			
			break;
			
		default:
			/* try another buffer */
			printf("weird nit state (%d) - skipping this buffer...\n",
			       nheader->nh_state);
			tryagain = 1;
			break;
		}
#endif
#else /* NETMON */
		nt = (struct netmonstr *)buf;
		if ( (old_seq != -1) && 
		    (diff = (int)((u_short)nt->seqno - (u_short)old_seq)) )    
			(void)printf("dropped %d netmon records\n", diff);
		old_seq = nt->seqno + (n/sizeof(struct netmonstr));
		/*
		 * NETMON:  Set pointers for this buffer
		 */
		*bp = buf;
		*endbuf = buf + n;
#ifdef NETMON_DEBUG
		{
			int i, j;
			
			nt = (struct netmonstr *)buf;
			for (i = 1; i < NM_MAXTRACED/2; i++, nt++) {
				printf("%d %d %d>%d %d",
				       nt->state, nt->seqno,
				       nt->code, nt->ql, nt->len);
				
				if(nt->len != 0) {
					for(j=0; j < nt->len; j++)
						printf(" %02x", nt->data[j]);
					printf("\n");
				}
				else
					printf("\n");
			}
		}
#endif
		tryagain = 0;
#endif
	} while (tryagain);
	}

void cleanup()
{
	(void) fflush(stdout);
	exit(0);
}


/*
  MAIN - option processing and calls getval()
 */
main(argc, argv)
	int	argc;
	char	**argv;
{
	int   	catch_inter(), process_packets();
		
#ifdef NIT_CATCH
		
#ifdef SUN3_NIT
	int	       	initsock();
#else	
	int	       	initstream();
#endif
		
#else /* NETMON */
	int	       	initnetmon();	/* NETMON */
#endif
	
	void	       	help_screen(), cleanup(), usage(), getval();
	int	       	loop = 1;
	char	       	*hostname;
	
	int		shours, smins, ssecs;
	unsigned long  	seconds;
	struct tm      	*stime;
	struct timeval 	*tps;
	struct timezone	*tzps;
	char		*sdate;
	char		*malloc();
	char		*ptr;

	ptr = (char *)rindex(argv[0], '/');
	if (ptr == NULL)
		ptr = argv[0];
	else
		ptr++;
	
       /*
	* Check to see what name we were called by and
	* do the right thing.
	*/
	if (strcmp("ositrace", ptr) == 0)
		tracetype |= OSITRACE;
	if (strcmp("pdutrace", ptr) == 0)
		tracetype |= OSITRACE;
	if (strcmp("tcptrace", ptr) == 0) 
		tracetype |= TCPTRACE;
		
	src_host = dst_host = 0;
	src_port = dst_port = 0;
	argc--;
	argv++;
		
	while (argc > 0) {
		if ((*argv)[0] == '-') {
			(*argv)++;
		}
		switch ((*argv)[0]) {
		case 'n':
			numbers= 1;
			break;
		case 'c':
			argc--;
			if (argc <= 0) {
				fprintf(stderr,"bad -c option\n");
				usage();
			}
			count = atoi(*++argv);
			while (isdigit(**argv) && **argv != '\0') {
				(*argv)++;
			}
			if (**argv != '\0' || count <= 0) {
				fprintf(stderr,"bad count\n");
				(void) usage();
		}
			break;
		case 'h':
			(void) help_screen();
			exit(0);
		break;
		case 'H':
			argc--;
			if (argc <= 0) {
				fprintf(stderr,"bad -H option\n");
				usage();
			}
			*++argv;
			if (!(tmp_hostent = gethostbyname(*argv))) {
				if ((src_host = dst_host = inet_addr(*argv)) == -1) {
					(void) fprintf(stderr,
					       "invalid host name:  %s\n(or nameserver is down)\n\n",*argv);
					exit (-2);
				} else {
					host = 1;
				}
			} else {
				src_host = dst_host = * (u_long *)tmp_hostent->h_addr;
				host = 1;
			}
			break;
		case 'p':
			argc--;
			if (argc <= 0) {
				fprintf(stderr,"bad -p option\n");
			usage();
			}
			port = 1;
			src_port = dst_port = atoi(*++argv);
			break;
		case 'r':
			if (!strcmp(*argv, "raw"))
				rawf = 1;
			break;
		case 'D':
			argc--;
			if (argc <= 0) {
				fprintf(stderr,"bad -D option\n");
				usage();
			}
			*++argv;
			if (!(tmp_hostent = gethostbyname(*argv))) {
				if ((dst_host = inet_addr(*argv)) == -1) {
					(void) fprintf(stderr,
						       "invalid host name:  %s\n(or nameserver is down)\n\n",*argv);
					exit (-2);
				}
			} else
				dst_host = * (u_long *)tmp_hostent->h_addr;
			break;
		case 'd':
			argc--;
			if (argc <= 0) {
				fprintf(stderr,"bad -d option\n");
				usage();
			}
			dst_port = atoi(*++argv);
			break;
		case 'S':
			argc--;
			if (argc <= 0) {
				fprintf(stderr,"bad -S option\n");
				usage();
			}
			*++argv;
			if (!(tmp_hostent = gethostbyname(*argv))) {
				if ((src_host = inet_addr(*argv)) == -1) {
					(void) fprintf(stderr,
						       "invalid host name:  %s\n(or nameserver is down)\n\n",*argv);
					exit (-2);
				}
			} else 
				src_host = * (u_long *)tmp_hostent->h_addr;
			break;
		case 's':
			argc--;
			if (argc <= 0) {
				fprintf(stderr,"bad -s option\n");
				usage();
			}
			src_port = atoi(*++argv);
			break;
		case 'z':
			argc--;
			if (argc <=0) {
				fprintf(stderr,"bad -z (sleep) option\n");
				usage();
			}
			*++argv;
			sscanf(*argv,"%d:%d:%d",&milhrs,&milmins,&milsecs);
			sleepf = 1;
			break;
		case 'q':
		case 'Q':
			quenchf = 1;
			break;
		case 'o':
		case 'O':
			tracetype |= OSITRACE;
			break;
		case 't':
		case 'T':
			tracetype |= TCPTRACE;
			break;
		case 'v':
			argc--;
			if (argc <= 0) {
				fprintf(stderr,"bad -s option\n");
				usage();
			}
			verbose = atoi(*++argv);
			break;
		default:
			fprintf(stderr,"bad option\n");
			(void) usage();
			break;
		}
		argv++;
		argc--;
	}
	
	if (!tracetype)		/* default to tcptrace */
		tracetype = TCPTRACE;

	if (tracetype & OSITRACE)
		snaplen = 90;  /* OSI needs a little larger snaplen */

	/* Initialize the hash table to 0 */
	
	connect_rec = NULL;
	for (loop = 0; loop <= TAB_SIZ - 1; loop++)
		hash_tab[loop] = NULL;
	
	/* sleep function doesn't work */
	/* must we sleep first ? */
	if (sleepf) {
		tps = (struct timeval *)malloc(sizeof(struct timeval));
		tzps = (struct timezone *)malloc(sizeof(struct timezone));
		sdate = malloc(30);
		gettimeofday(tps,tzps);
		seconds = tps->tv_sec;
		sdate = ctime(&seconds);
		sscanf(sdate,"%*s %*s %*d %d:%d:%d %*d",&shours,&smins,&ssecs);
		sleepsecs = (milhrs-shours)*3600+(milmins-smins)*60+milsecs-ssecs;
		if (sleepsecs>0)
			sleep(sleepsecs);
	}
	
	wbuf = malloc(8192);
	setbuf(stdout,wbuf);
	
#ifdef NIT_CATCH
	
#ifdef SUN3_NIT
	fd = initsock();
#else  /* SUN4_NIT */
	fd = initstream();
#endif
	
#else /* NETMON */
	fd = initnetmon();	/* NETMON */
	kif_init();
#endif
	
#ifdef NIT_CATCH
	/* find this machine's name to catch loop closing FINs */
	
	me = 0;
	hostname = (char *) malloc (BUFSIZ);
	if (gethostname(hostname, BUFSIZ)) {
		(void) fprintf(stderr,"can't find this machine's name\n\n");
	} else if ((tmp_hostent = gethostbyname(hostname)) == NULL) {
		(void) fprintf(stderr,"gethostbyname failed on %s, me = 0\n\n",
			       hostname);
		me = 0;
	} else
		me = *(u_long *)tmp_hostent->h_addr;
#endif
	
	(void) signal(SIGINT, cleanup);

	process_packets();
	(void) cleanup();
	exit(0);
}


/*
  Get data from the network interface, call appropriate protocol
  routines to process data.
 */

process_packets()
{
#ifdef SUN4_NIT
	struct nit4_hdr *nheader;
#else
	struct	nit_hdr *nheader;
#endif
	struct  tm *localtime();
	struct	tm *timestamp;
	long	tot_sec, u_sec;
	struct	ether_header *eheader;
	unsigned char *bp, *endbuf, *start_of_protocol;
	int push_count = 0;
	int protocol;
	int loop = 0;
	static	int hdrsiz = 0;

	/* make sure we've got packets in the buffer */
	
	if (!count) {
		push_count++;
		count = 10000;
	}
	while (count > 0) {
		
		/* get a buffer from the ethernet tap or NETMON */
		
		get_buf(fd, &bp, &endbuf);
		
	       /*
		 Loop for each packet contained in the buffer.  At the
		 top of this loop, bp should point to the beginning of
		 the next packet to examine.
		*/
		do {
			/*
			 * this code all works off of bp and endbuf.
			 * bp points to the beginning of the current packet.
			 * endbuf points to the end of the entire buffer.
			 * they must be set correctly at this point.
			 */

			if (bp >= endbuf)
				continue;
			
#ifdef NIT_CATCH
			/*
			  Set up NIT pointers and check various NIT
			  headers.  We assume at this point that
			  bp is pointing to the start of the next
			  packet in the read bufer.
			 */
#ifdef SUN4_NIT
			nheader = (struct nit4_hdr *) bp;
			/*
			  Save count of dropped packets.
			  This value is cumlative
			 */
			numdropped = nheader->DROPPED;	

		       /*
			 Get a pointer to the ethernet header portion of
			 this packet.
			*/
			eheader = (struct ether_header *)
				  (bp + sizeof(struct nit4_hdr));
#ifdef NIT4_DEBUG
			dump_nit4_header(nheader);
			dump_ether_header(eheader);
#endif /* NIT4_DEBUG */
			/*
			  Get bp pointing to the start of the next NIT packet.
			 */
			bp += nheader->TOTLEN;
#else /* SUN3_NIT */   
			nheader = (struct nit_hdr *) bp;
		       /*
			* Scan pkts until we've got something 
			* worth looking at.
			*/
			while (nheader->nh_state != NIT_CATCH && bp < endbuf) {
				if (nheader->nh_state == 4) {
					/* BAD NEWS */
					(void)printf("Dropped Packets due to full buffer - ");
					(void)printf("number dropped = %d\n",
						     nheader->nh_un.dropped);
					numdropped += nheader->nh_un.dropped;
				}
				else
					(void) printf("strange packet : state == %d\n",
						      nheader->nh_state);
				/* skip to next packet */
		
				bp += sizeof(struct nit_hdr);
				nheader = (struct nit_hdr *) bp;
			}
			fflush(stdout);
	
			if (bp >= endbuf)
				continue;
	
			hdrsiz = nheader->nh_datalen;

			eheader = (struct ether_header *)
				(bp + sizeof(struct nit_hdr));
	
			/*
			 * increment ptr to start of following int boundary,
			 * past the previous packet.  (cons's up the proper
			 * number of bytes to increment and clears the bottom
			 * log-base-2(sizeof(int)) bits.)
			 * This is necessary due to observable behavior by NIT
			 * and therefore is NIT-dependent. This strategy seems
			 *  to work...
vv			 */
			bp += ((sizeof(struct nit_hdr) + hdrsiz
				 + sizeof(int) - 1) & ~(sizeof (int) - 1));
#endif /* SUN3_NIT */	

#endif	/* NIT_CATCH */
#ifdef NETMON
	
			nt = (struct netmonstr *)bp;
			
			bp += sizeof(struct netmonstr);  /* point to next record */

			if(nt->info) 
				sprintf(netmonif, "%s",
					(char *)kif_name(nt->info));
			else
				netmonif[0] = NULL;

			netmoncode = nt->code;
				
			switch (nt->loc) {
			case IP_IN:

			case IP_OUT:
			case CLNP_IN:
			case CLNP_OUT:
				netmonloc = nt->loc;
				break;
			default:
				continue;
			}
			/*
			 * NETMON: analyze only records with header info. 
			 */
			if (nt->len == 0) {
				continue;
			}
#endif /* NETMON */

			/*
			 * Bump up counter for total number of pkts examined
			 */
			numseen++;

			/* get the timestamp */
#ifdef NIT_CATCH
	
#ifdef SUN4_NIT
			timestamp	= localtime(&(nheader->SECS));
			tot_sec		= nheader->SECS;
			u_sec		= nheader->USECS;
#else	
			timestamp = localtime(&(nheader->nh_timestamp.tv_sec));
			tot_sec = nheader->nh_timestamp.tv_sec;
			u_sec = nheader->nh_timestamp.tv_usec;
#endif /* SUN4_NIT */
			start_of_protocol = (unsigned char *) eheader
					    + sizeof(struct ether_header);

			protocol = eheader->ether_type;
#endif

#ifdef NETMON
			/*
			 * NETMON:  correct BSD4.3 usec counts from microtime.
			 */
	
			if(nt->ts.tv_usec == 1000000) {
				nt->ts.tv_sec++;
				nt->ts.tv_usec = 0;
			}
	
			timestamp 	= localtime(&(nt->ts.tv_sec));
			tot_sec 	= nt->ts.tv_sec;
			u_sec 		= nt->ts.tv_usec;
			start_of_protocol = (unsigned char *) nt +
				sizeof(struct netmonstr) - (NM_MAXDATA + 2);
/*			dump_buf(start_of_protocol, NM_MAXDATA);*/
			protocol = 81;
#endif /* NETMON */	

			switch (protocol) {

			case ETHERTYPE_IP:
				/*
				 * Check to see if we are tracing TCP packets.
				 * If so, process this packet.
				 */
				if (tracetype & TCPTRACE)
					process_ip_packet(src_host,
						dst_host, host, src_port,
						dst_port, port,
					        timestamp, tot_sec, u_sec,
						start_of_protocol);
				break;
			default:
				/*
				 * Check to see if we are tracing OSI packets.
				 * Also check to see if the ethernet type/len
				 * field is <= 1500, this is what SUNLINK OSI
				 * sticks there for now.
				 */
				if (   (tracetype & OSITRACE)
				    && (protocol <= 1500)) {
					fflush(stdout);
#ifndef NETMON
					start_of_protocol += 3;
#endif
					process_osi_packet(start_of_protocol,
							   timestamp,
							   tot_sec,
							   u_sec);
				}
			} /* end case */
			count--;
			if (push_count)
				count = 10000;
		} while ((bp < endbuf) && (count > 0));
	} /* end while */
}

