/*

Copyright (c) 1989, the MITRE Corporation. All rights reserved.
This software was developed by the MITRE Corporation, McLean VA.
It is distributed free of charge with the following conditions:

        1.  Redistribution and use in source and binary forms are
	    permitted provided that this copyright notice be duplicated
	    in all such forms.

        2.  MITRE will be credited should the software be used in a
            product or written about in any publications.  However,
            MITRE will not be referenced as the source in advertisements.

        3.  MITRE assumes no legal responsibility for source code and
            its subsequent use.  No warranty is expressed or implied.

        4.  If any bugs or problems are found, then they should be
            reported back to MITRE.

*/

#include <stdio.h>
#include <netdb.h>
#include <ctype.h>
#include <string.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/errno.h>
#include <signal.h>

#include <net/if.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/if_ether.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/ip_icmp.h>

/*
  Important defines:
 
 	SUN3_NIT - causes code to be compiled for a SunOS 3 operating system
 	SUN4_NIT - causes code to be compiled for a SunOS 4 operating system
*/

/*#define NETMON 	1*/
#define  SUN4_NIT 	1

/*
  Compilation Flag for debugging the SunOS 4.0 version of tcptrace
*/ 
   
/*#define  NIT4_DEBUG 1*/
/*#define  PERFDEBUG*/

#if SUN3_NIT || SUN4_NIT

#include <net/nit.h>

#ifdef 	SUN4_NIT

#include <net/nit_pf.h>
#include <net/nit_buf.h>
#include <net/nit_if.h>
#include <fcntl.h>
#include <stropts.h>

#define NIT_DEV "/dev/nit"
/*
  Define the number of 1k chunks to allocate for the NIT buffer.

  The default, 30k, requires a modified kernel.  The kernel file
  param.c contains the stream buffer allocation sizes.  

*/
#define NCHUNKS 30

#define READSIZE 32 * 1024

struct nit4_hdr {
	struct nit_bufhdr 	buf_hdr;
	struct nit_iftime 	time_hdr;
	struct nit_ifdrops 	drop_hdr;
	struct nit_iflen  	iflen_hdr;
};

/*
  Define some NIT header fields
 */
#define MSGLEN  buf_hdr.nhb_msglen
#define TOTLEN  buf_hdr.nhb_totlen
#define SECS    time_hdr.nh_timestamp.tv_sec
#define USECS   time_hdr.nh_timestamp.tv_usec
#define DROPPED drop_hdr.nh_drops
#define PKTLEN  iflen_hdr.nh_pktlen

#else	/* SUN3_NIT */

/*
  
  Define the number of 1k chunks to allocate for the NIT buffer.
  30k should be no problem for the standard sun3 kernel?
  
*/
#define NCHUNKS 30

#endif 

#else /* NETMON */

#include <sys/inode.h>
#include <sys/file.h>

#define KERNEL
#define NETMON_KERNEL
#include <net/netmon.h>

#endif /* NETMON */


/* defines for fields in tcp_header */

#define tcp_s_port	th_sport;		/* source port */
#define tcp_d_port	th_dport;		/* destination port */
#define tcp_seqno	th_seq;			/* sequence number */
#define tcp_ack		tcp_seq	th_ack;		/* acknowledgement number */

/* Configuration parameters */

					/* # seconds before a read times out */
#define READ_TIMEOUT	10		/* probably longer than it has to be */

/*
 * connection record for use in the hash table.
 */
struct	conx_rec {
	u_long	th_seq;
        u_long  th_ack;
	char	*hostname1,	
		*hostname2;
	char	*dot_addr1,
		*dot_addr2;
	struct in_addr ip_addr1,
	    	ip_addr2;
	u_long	fin;
	struct	tm *timestmp;
	long	tot_sec;
	u_short	th_port1,
	    	th_port2;
	short	closed;
	int	packets;
	int	extended;
	u_char	flags;
	struct	conx_rec *next;
};

/*
 * Valid flags values
 */
#define TCP		0x1
#define OSI		0x2
#define VISITED		0x4
#define EXTENDED	0x8

#define	TAB_SIZ		131	/* lowest prime > 128 for hashing */
#define	MAX_LOOPS	200	/* # loops before garbage collection (g.c.) */
#define	MAX_RECS	129	/* max # records allocated before g.c. */

#ifdef NIT_CATCH
#define CHUNKSIZE	(NCHUNKS * 1024)
#define SNAPLEN		(sizeof(struct ether_header) \
			 + sizeof(struct ip) + sizeof(struct tcphdr))
#else /* NETMON */
#define CHUNKSIZE       (NM_MAXTRACED * sizeof(struct netmonstr))

#endif

#define NETBUFSIZ	(32 * 1024)
