/*

Copyright (c) 1989, the MITRE Corporation. All rights reserved.
This software was developed by the MITRE Corporation, McLean VA.
It is distributed free of charge with the following conditions:

        1.  Redistribution and use in source and binary forms are
	    permitted provided that this copyright notice be duplicated
	    in all such forms.

        2.  MITRE will be credited should the software be used in a
            product or written about in any publications.  However,
            MITRE will not be referenced as the source in advertisements.

        3.  MITRE assumes no legal responsibility for source code and
            its subsequent use.  No warranty is expressed or implied.

        4.  If any bugs or problems are found, then they should be
            reported back to MITRE.

*/

#include "trace.h"
	
/*
 * USAGE - called on bad usage
 */
	
void usage()
{
	(void) fprintf(stderr, "usage:  tcptrace [-n | -[DHS] hostname| -c count]");
	(void) fprintf(stderr, " | -[psd] portnumber\n");
	(void) fprintf(stderr, "type: tcptrace -h   for help.\n");
	exit(-1);
}

/*
 * help screen printing.
 */	
void help_screen()
{
	printf("Tcptrace logs to stdout a record of packets seen on the net.\n");
	printf("The options given to tcptrace are:\n\n");
	printf("-c count		How many packets to trace before termination.\n");
	printf("-n			Print dot addresses instead of host names.\n");
	printf("-h			Print this screen.\n");
	printf("-D dsthost              specify destination host.\n");
	printf("-H hostname		specify host as source or destination\n");
	printf("-S srchost              specify source host\n");
	printf("-p <portnumber>         specify port number - source or destination\n");
	printf("-s <portnumber>	        specify source port number.\n");
	printf("-d <portnumber>         specify destination port number.\n");
	printf("-t			specify TCP packet tracing (default)\n");
	printf("-o			specify OSI packet tracing\n");
	printf("	Use both -t and -o to trace both TCP and OSI\n");
	printf("\n\n");	       
	printf("problems:  tcptrace uses gethostby<name, addr> which might poll a name server.\n");
	printf("If the name server is out to lunch, this will fail, and the program will crash.\n");
	printf("The count option is not exact, but should be within a few or the desired number.\n");
}



#ifdef SUN4_NIT
	
dump_nit4_header(nheader)
	struct nit4_hdr *nheader;
{
	printf("MSGLEN: %ld\tTOTLEN: %ld\tDROPPED: %ld\tPKTLEN: %ld\n",
	       nheader->MSGLEN, nheader->TOTLEN, nheader->DROPPED,
	       nheader->PKTLEN);
}


dump_ether_header(eheader)
	struct ether_header *eheader;
{
	printf("DST ADDR: %x.%x.%x.%x.%x.%x\tSRC ADDR: %x.%x.%x.%x.%x.%x\tTYPE/LEN: %x\n\n",
	       (u_char) eheader->ether_dhost.ether_addr_octet[0],
	       (u_char) eheader->ether_dhost.ether_addr_octet[1],
	       (u_char) eheader->ether_dhost.ether_addr_octet[2],
	       (u_char) eheader->ether_dhost.ether_addr_octet[3],
	       (u_char) eheader->ether_dhost.ether_addr_octet[4],
	       (u_char) eheader->ether_dhost.ether_addr_octet[5],
	       (u_char) eheader->ether_shost.ether_addr_octet[0],
	       (u_char) eheader->ether_shost.ether_addr_octet[1],
	       (u_char) eheader->ether_shost.ether_addr_octet[2],
	       (u_char) eheader->ether_shost.ether_addr_octet[3],
	       (u_char) eheader->ether_shost.ether_addr_octet[4],
	       (u_char) eheader->ether_shost.ether_addr_octet[5],
	       (u_short) eheader->ether_type);
}

#endif /* SUN4_NIT */

dump_buf(buf, len)
	char *buf;
	int len;
{
	int i, j;
	
	i=j=0;
	while (i<len) {
		printf("%3d:[%2x] ", i, (u_char) buf[i]);
		i++; j++;
		if (j == 8) {
			printf("\n");
			j=0;
		}
	}
	printf("\n");
}

